@@ -1,20 +1,18 @@
-use google_cloud_googleapis::{Code, Status};
-use tonic::{Response, Streaming};
+use google_cloud_gax::cancel::CancellationToken;
+use std::time::Duration;
 
-use google_cloud_gax::call_option::{Backoff, BackoffRetrySettings, BackoffRetryer};
-use google_cloud_gax::invoke::invoke_reuse;
-use google_cloud_gax::util::create_request;
+use google_cloud_gax::conn::Channel;
+use google_cloud_gax::create_request;
+use google_cloud_gax::grpc::{Code, Response, Status, Streaming};
+use google_cloud_gax::retry::{invoke_fn, RetrySetting};
 use google_cloud_googleapis::spanner::v1 as internal;
 use google_cloud_googleapis::spanner::v1::spanner_client::SpannerClient;
 use google_cloud_googleapis::spanner::v1::{
-    BatchCreateSessionsRequest, BatchCreateSessionsResponse, BeginTransactionRequest,
-    CommitRequest, CommitResponse, CreateSessionRequest, DeleteSessionRequest,
-    ExecuteBatchDmlRequest, ExecuteBatchDmlResponse, ExecuteSqlRequest, GetSessionRequest,
-    ListSessionsRequest, ListSessionsResponse, PartialResultSet, PartitionQueryRequest,
-    PartitionReadRequest, PartitionResponse, ReadRequest, ResultSet, RollbackRequest, Session,
-    Transaction,
+    BatchCreateSessionsRequest, BatchCreateSessionsResponse, BeginTransactionRequest, CommitRequest, CommitResponse,
+    CreateSessionRequest, DeleteSessionRequest, ExecuteBatchDmlRequest, ExecuteBatchDmlResponse, ExecuteSqlRequest,
+    GetSessionRequest, ListSessionsRequest, ListSessionsResponse, PartialResultSet, PartitionQueryRequest,
+    PartitionReadRequest, PartitionResponse, ReadRequest, ResultSet, RollbackRequest, Session, Transaction,
 };
-use google_cloud_grpc::conn::Channel;
 
 pub(crate) fn ping_query_request(session_name: impl Into<String>) -> internal::ExecuteSqlRequest {
     internal::ExecuteSqlRequest {
@@ -32,13 +30,13 @@ pub(crate) fn ping_query_request(session_name: impl Into<String>) -> internal::E
     }
 }
 
-fn default_setting() -> BackoffRetrySettings {
-    BackoffRetrySettings {
-        retryer: BackoffRetryer {
-            backoff: Backoff::default(),
-            //handle gRPC stream error (Status { code: Unknown, message: "transport error", source: None })
-            codes: vec![Code::Unavailable, Code::Unknown],
-        },
+fn default_setting() -> RetrySetting {
+    RetrySetting {
+        from_millis: 50,
+        max_delay: Some(Duration::from_secs(10)),
+        factor: 1u64,
+        take: 20,
+        codes: vec![Code::Unavailable, Code::Unknown],
     }
 }
 
@@ -53,14 +51,6 @@ impl Client {
         Client { inner }
     }
 
-    /// merge call setting
-    fn get_call_setting(call_setting: Option<BackoffRetrySettings>) -> BackoffRetrySettings {
-        match call_setting {
-            Some(s) => s,
-            None => default_setting(),
-        }
-    }
-
     /// create_session creates a new session. A session can be used to perform
     /// transactions that read and/or modify data in a Cloud Spanner database.
     /// Sessions are meant to be reused for many consecutive
@@ -83,11 +73,14 @@ impl Client {
     pub async fn create_session(
         &mut self,
         req: CreateSessionRequest,
-        opt: Option<BackoffRetrySettings>,
+        cancel: Option<CancellationToken>,
+        retry: Option<RetrySetting>,
     ) -> Result<Response<Session>, Status> {
-        let mut setting = Client::get_call_setting(opt);
+        let setting = retry.unwrap_or(default_setting());
         let database = &req.database;
-        return invoke_reuse(
+        return invoke_fn(
+            cancel,
+            Some(setting),
             |spanner_client| async {
                 let request = create_request(format!("database={}", database), req.clone());
                 spanner_client
@@ -96,7 +89,6 @@ impl Client {
                     .map_err(|e| (e.into(), spanner_client))
             },
             &mut self.inner,
-            &mut setting,
         )
         .await;
     }
@@ -108,11 +100,14 @@ impl Client {
     pub async fn batch_create_sessions(
         &mut self,
         req: BatchCreateSessionsRequest,
-        opt: Option<BackoffRetrySettings>,
-    ) -> Result<tonic::Response<BatchCreateSessionsResponse>, Status> {
-        let mut setting = Client::get_call_setting(opt);
+        cancel: Option<CancellationToken>,
+        retry: Option<RetrySetting>,
+    ) -> Result<Response<BatchCreateSessionsResponse>, Status> {
+        let setting = retry.unwrap_or(default_setting());
         let database = &req.database;
-        return invoke_reuse(
+        return invoke_fn(
+            cancel,
+            Some(setting),
             |spanner_client| async {
                 let request = create_request(format!("database={}", database), req.clone());
                 spanner_client
@@ -121,7 +116,6 @@ impl Client {
                     .map_err(|e| (e.into(), spanner_client))
             },
             &mut self.inner,
-            &mut setting,
         )
         .await;
     }
@@ -132,11 +126,14 @@ impl Client {
     pub async fn get_session(
         &mut self,
         req: GetSessionRequest,
-        opt: Option<BackoffRetrySettings>,
+        cancel: Option<CancellationToken>,
+        retry: Option<RetrySetting>,
     ) -> Result<Response<Session>, Status> {
-        let mut setting = Client::get_call_setting(opt);
+        let setting = retry.unwrap_or(default_setting());
         let name = &req.name;
-        return invoke_reuse(
+        return invoke_fn(
+            cancel,
+            Some(setting),
             |spanner_client| async {
                 let request = create_request(format!("name={}", name), req.clone());
                 spanner_client
@@ -145,7 +142,6 @@ impl Client {
                     .map_err(|e| (e.into(), spanner_client))
             },
             &mut self.inner,
-            &mut setting,
         )
         .await;
     }
@@ -154,11 +150,14 @@ impl Client {
     pub async fn list_sessions(
         &mut self,
         req: ListSessionsRequest,
-        opt: Option<BackoffRetrySettings>,
+        cancel: Option<CancellationToken>,
+        retry: Option<RetrySetting>,
     ) -> Result<Response<ListSessionsResponse>, Status> {
-        let mut setting = Client::get_call_setting(opt);
+        let setting = retry.unwrap_or(default_setting());
         let database = &req.database;
-        return invoke_reuse(
+        return invoke_fn(
+            cancel,
+            Some(setting),
             |spanner_client| async {
                 let request = create_request(format!("database={}", database), req.clone());
                 spanner_client
@@ -167,7 +166,6 @@ impl Client {
                     .map_err(|e| (e.into(), spanner_client))
             },
             &mut self.inner,
-            &mut setting,
         )
         .await;
     }
@@ -178,11 +176,14 @@ impl Client {
     pub async fn delete_session(
         &mut self,
         req: DeleteSessionRequest,
-        opt: Option<BackoffRetrySettings>,
+        cancel: Option<CancellationToken>,
+        retry: Option<RetrySetting>,
     ) -> Result<Response<()>, Status> {
-        let mut setting = Client::get_call_setting(opt);
+        let setting = retry.unwrap_or(default_setting());
         let name = &req.name;
-        return invoke_reuse(
+        return invoke_fn(
+            cancel,
+            Some(setting),
             |spanner_client| async {
                 let request = create_request(format!("name={}", name), req.clone());
                 spanner_client
@@ -191,7 +192,6 @@ impl Client {
                     .map_err(|e| (e.into(), spanner_client))
             },
             &mut self.inner,
-            &mut setting,
         )
         .await;
     }
@@ -210,11 +210,14 @@ impl Client {
     pub async fn execute_sql(
         &mut self,
         req: ExecuteSqlRequest,
-        opt: Option<BackoffRetrySettings>,
+        cancel: Option<CancellationToken>,
+        retry: Option<RetrySetting>,
     ) -> Result<Response<ResultSet>, Status> {
-        let mut setting = Client::get_call_setting(opt);
+        let setting = retry.unwrap_or(default_setting());
         let session = &req.session;
-        return invoke_reuse(
+        return invoke_fn(
+            cancel,
+            Some(setting),
             |spanner_client| async {
                 let request = create_request(format!("session={}", session), req.clone());
                 spanner_client
@@ -223,7 +226,6 @@ impl Client {
                     .map_err(|e| (e.into(), spanner_client))
             },
             &mut self.inner,
-            &mut setting,
         )
         .await;
     }
@@ -236,11 +238,14 @@ impl Client {
     pub async fn execute_streaming_sql(
         &mut self,
         req: ExecuteSqlRequest,
-        opt: Option<BackoffRetrySettings>,
+        cancel: Option<CancellationToken>,
+        retry: Option<RetrySetting>,
     ) -> Result<Response<Streaming<PartialResultSet>>, Status> {
-        let mut setting = Client::get_call_setting(opt);
+        let setting = retry.unwrap_or(default_setting());
         let session = &req.session;
-        return invoke_reuse(
+        return invoke_fn(
+            cancel,
+            Some(setting),
             |spanner_client| async {
                 let request = create_request(format!("session={}", session), req.clone());
                 spanner_client
@@ -249,7 +254,6 @@ impl Client {
                     .map_err(|e| (e.into(), spanner_client))
             },
             &mut self.inner,
-            &mut setting,
         )
         .await;
     }
@@ -268,34 +272,33 @@ impl Client {
     pub async fn execute_batch_dml(
         &mut self,
         req: ExecuteBatchDmlRequest,
-        opt: Option<BackoffRetrySettings>,
+        cancel: Option<CancellationToken>,
+        retry: Option<RetrySetting>,
     ) -> Result<Response<ExecuteBatchDmlResponse>, Status> {
-        let mut setting = Client::get_call_setting(opt);
+        let setting = retry.unwrap_or(default_setting());
         let session = &req.session;
-        return invoke_reuse(
+        return invoke_fn(
+            cancel,
+            Some(setting),
             |spanner_client| async {
                 let request = create_request(format!("session={}", session), req.clone());
                 let result = spanner_client.execute_batch_dml(request).await;
                 match result {
                     Ok(response) => match response.get_ref().status.as_ref() {
                         Some(s) => {
-                            let tonic_code = tonic::Code::from(s.code);
-                            if tonic_code == tonic::Code::Ok {
+                            let code = Code::from(s.code);
+                            if code == Code::Ok {
                                 Ok(response)
                             } else {
-                                Err((
-                                    tonic::Status::new(tonic_code, s.message.to_string()).into(),
-                                    spanner_client,
-                                ))
+                                Err((Status::new(code, s.message.to_string()), spanner_client))
                             }
                         }
                         None => Ok(response),
                     },
-                    Err(err) => Err((err.into(), spanner_client)),
+                    Err(err) => Err((err, spanner_client)),
                 }
             },
             &mut self.inner,
-            &mut setting,
         )
         .await;
     }
@@ -316,11 +319,14 @@ impl Client {
     pub async fn read(
         &mut self,
         req: ReadRequest,
-        opt: Option<BackoffRetrySettings>,
+        cancel: Option<CancellationToken>,
+        retry: Option<RetrySetting>,
     ) -> Result<Response<ResultSet>, Status> {
-        let mut setting = Client::get_call_setting(opt);
+        let setting = retry.unwrap_or(default_setting());
         let session = &req.session;
-        return invoke_reuse(
+        return invoke_fn(
+            cancel,
+            Some(setting),
             |spanner_client| async {
                 let request = create_request(format!("session={}", session), req.clone());
                 spanner_client
@@ -329,7 +335,6 @@ impl Client {
                     .map_err(|e| (e.into(), spanner_client))
             },
             &mut self.inner,
-            &mut setting,
         )
         .await;
     }
@@ -342,11 +347,14 @@ impl Client {
     pub async fn streaming_read(
         &mut self,
         req: ReadRequest,
-        opt: Option<BackoffRetrySettings>,
+        cancel: Option<CancellationToken>,
+        retry: Option<RetrySetting>,
     ) -> Result<Response<Streaming<PartialResultSet>>, Status> {
-        let mut setting = Client::get_call_setting(opt);
+        let setting = retry.unwrap_or(default_setting());
         let session = &req.session;
-        return invoke_reuse(
+        return invoke_fn(
+            cancel,
+            Some(setting),
             |spanner_client| async {
                 let request = create_request(format!("session={}", session), req.clone());
                 spanner_client
@@ -355,7 +363,6 @@ impl Client {
                     .map_err(|e| (e.into(), spanner_client))
             },
             &mut self.inner,
-            &mut setting,
         )
         .await;
     }
@@ -367,11 +374,14 @@ impl Client {
     pub async fn begin_transaction(
         &mut self,
         req: BeginTransactionRequest,
-        opt: Option<BackoffRetrySettings>,
+        cancel: Option<CancellationToken>,
+        retry: Option<RetrySetting>,
     ) -> Result<Response<Transaction>, Status> {
-        let mut setting = Client::get_call_setting(opt);
+        let setting = retry.unwrap_or(default_setting());
         let session = &req.session;
-        return invoke_reuse(
+        return invoke_fn(
+            cancel,
+            Some(setting),
             |spanner_client| async {
                 let request = create_request(format!("session={}", session), req.clone());
                 spanner_client
@@ -380,7 +390,6 @@ impl Client {
                     .map_err(|e| (e.into(), spanner_client))
             },
             &mut self.inner,
-            &mut setting,
         )
         .await;
     }
@@ -402,11 +411,14 @@ impl Client {
     pub async fn commit(
         &mut self,
         req: CommitRequest,
-        opt: Option<BackoffRetrySettings>,
+        cancel: Option<CancellationToken>,
+        retry: Option<RetrySetting>,
     ) -> Result<Response<CommitResponse>, Status> {
-        let mut setting = Client::get_call_setting(opt);
+        let setting = retry.unwrap_or(default_setting());
         let session = &req.session;
-        return invoke_reuse(
+        return invoke_fn(
+            cancel,
+            Some(setting),
             |spanner_client| async {
                 let request = create_request(format!("session={}", session), req.clone());
                 spanner_client
@@ -415,7 +427,6 @@ impl Client {
                     .map_err(|e| (e.into(), spanner_client))
             },
             &mut self.inner,
-            &mut setting,
         )
         .await;
     }
@@ -431,11 +442,14 @@ impl Client {
     pub async fn rollback(
         &mut self,
         req: RollbackRequest,
-        opt: Option<BackoffRetrySettings>,
+        cancel: Option<CancellationToken>,
+        retry: Option<RetrySetting>,
     ) -> Result<Response<()>, Status> {
-        let mut setting = Client::get_call_setting(opt);
+        let setting = retry.unwrap_or(default_setting());
         let session = &req.session;
-        return invoke_reuse(
+        return invoke_fn(
+            cancel,
+            Some(setting),
             |spanner_client| async {
                 let request = create_request(format!("session={}", session), req.clone());
                 spanner_client
@@ -444,7 +458,6 @@ impl Client {
                     .map_err(|e| (e.into(), spanner_client))
             },
             &mut self.inner,
-            &mut setting,
         )
         .await;
     }
@@ -463,11 +476,14 @@ impl Client {
     pub async fn partition_query(
         &mut self,
         req: PartitionQueryRequest,
-        opt: Option<BackoffRetrySettings>,
+        cancel: Option<CancellationToken>,
+        retry: Option<RetrySetting>,
     ) -> Result<Response<PartitionResponse>, Status> {
-        let mut setting = Client::get_call_setting(opt);
+        let setting = retry.unwrap_or(default_setting());
         let session = &req.session;
-        return invoke_reuse(
+        return invoke_fn(
+            cancel,
+            Some(setting),
             |spanner_client| async {
                 let request = create_request(format!("session={}", session), req.clone());
                 spanner_client
@@ -476,7 +492,6 @@ impl Client {
                     .map_err(|e| (e.into(), spanner_client))
             },
             &mut self.inner,
-            &mut setting,
         )
         .await;
     }
@@ -497,11 +512,14 @@ impl Client {
     pub async fn partition_read(
         &mut self,
         req: PartitionReadRequest,
-        opt: Option<BackoffRetrySettings>,
+        cancel: Option<CancellationToken>,
+        retry: Option<RetrySetting>,
     ) -> Result<Response<PartitionResponse>, Status> {
-        let mut setting = Client::get_call_setting(opt);
+        let setting = retry.unwrap_or(default_setting());
         let session = &req.session;
-        return invoke_reuse(
+        return invoke_fn(
+            cancel,
+            Some(setting),
             |spanner_client| async {
                 let request = create_request(format!("session={}", session), req.clone());
                 spanner_client
@@ -510,7 +528,6 @@ impl Client {
                     .map_err(|e| (e.into(), spanner_client))
             },
             &mut self.inner,
-            &mut setting,
         )
         .await;
     }
